import os
import subprocess
import re
import consul
import registry

# Do not assign a local IP to the node
# DOCKER_FIX trick to avoid this issue:
# https://github.com/docker/docker/issues/14203
DOCKER_RUN_OPTS = ('--net="none" '
                   '-v /root/.ssh/authorized_keys:/root/.ssh/authorized_keys '
                   '-d -ti -e DOCKER_FIX=""')


class Volume(object):
    """Represents a Docker Volume"""
    def __init__(self, origin=None, destination=None, mode=None):
        self.origin = origin
        self.destination = destination
        self.mode = mode


class Network(object):
    """Represents a Docker Network Device"""
    def __init__(self, device=None, address=None,
                 bridge=None, netmask=None, gateway=None):
        self.device = device
        self.address = address
        self.bridge = bridge
        self.netmask = netmask
        self.gateway = gateway


def run(nodedn):
    """Run a given container

    This command gets the info needed to launch the container from the registry.
    Information retrieved from the registry:

    Node object:

    - name: name to give to the docker container
    - clustername: name of the cluster/service to which this docker belongs
    - docker_image
    - docker_opts
    - id: docker id
    - disks: Disk object list (see below)
    - networks: Network object list (see below)
    - tags: ('master', 'yarn')
    - status: pending, running, failed, stopped
    - host: docker engine where the container is running
    - port: main service port, e.g. 22
    - check_ports: list of ports to check that the container is alive

    Network object (registry.Network object):

    - device
    - address
    - bridge
    - netmask
    - gateway

    Volume object (registry.Disk object):

    - origin
    - destination
    - mode
    """
    node = registry.Node(nodedn)

    container_name = '{0}-{1}'.format(node.clustername, node.name)
    container_image = node.docker_image
    docker_opts = node.docker_opts
    service = node.clustername
    tags = node.tags
    disks = node.disks
    networks = node.networks

    opts = generate_docker_opts(docker_opts)
    volumes = generate_volume_opts(disks)

    _cmd('docker run {opts} {volumes} -h {name} --name {name} {image}'.format(
        name=container_name, opts=opts, volumes=volumes, image=container_image))
    add_network_connectivity(container_name, networks)
    register_in_consul(container_name, service, networks[0].address,
                       tags=tags, check_ports='SSH')


def generate_volume_opts(volumes):
    volume_opts = ''
    for volume in volumes:
        if not os.path.exists(volume.origin):
            os.mkdir(volume.origin)
        volume_opts += '-v {}:{}:{} '.format(volume.origin, volume.destination, volume.mode)
    return volume_opts


def generate_docker_opts(extra_opts):
    opts = DOCKER_RUN_OPTS
    opts += ' ' + extra_opts + ' '
    return opts


def add_network_connectivity(container_name, networks):
    """Adds the public networks interfaces to the given container"""
    #put('files/bin/pipework', '/tmp/pipework')
    #run('chmod u+x /tmp/pipework')
    for network in networks:
        add_network_interface(container_name, network)


def add_network_interface(container_name, network):
    """Adds one network interface using pipework

    TODO: If no address is specified obtain one from the network service
    """
    device = network.device
    bridge = network.bridge
    address = network.address
    netmask = network.netmask
    gateway = network.gateway

    if gateway and re.search(r'^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$', gateway):
        return _cmd(
            'pipework {bridge} -i {device} {name} {ip}/{mask}@{gateway}'
            .format(bridge=bridge, device=device, name=container_name,
                    ip=address,
                    mask=netmask,
                    gateway=gateway))
    else:
        return _cmd(
            'pipework {bridge} -i {device} {name} {ip}/{mask}'
            .format(bridge=bridge, device=device, name=container_name,
                    ip=address,
                    mask=netmask))

    #subprocess.call('pipework virbrSTORAGE -i eth0 {name} {ip}/{mask}'
    #    .format(name=container_name, ip=networks['storage']['address'],
    #            mask=networks['storage']['netmask']))
    #subprocess.call('pipework virbrPRIVATE -i eth1 {name} {ip}/{mask}@{gateway}'
    #    .format(name=container_name, ip=networks['private']['address'],
    #            mask=networks['private']['netmask'],
    #            gateway=networks['gateway']))


def register_in_consul(container_name, service_name, address,
                       tags=None, port=None, check_ports=None):
    """Register the docker container in consul service discovery"""
    sd = consul.Client()
    if check_ports:
        checks = generate_checks(container_name, address, check_ports)
    sd.register(container_name, service_name, address, tags=tags, port=port, check=checks)


def generate_checks(container, address, check_ports):
    """Generates the check dictionary to pass to consul of the form {'checks': []}"""
    checks = {}
    checks['checks'] = []
    for p in check_ports:
        checks['checks'].append(
            {'id': '{}-port{}'.format(container, p),
             'name': 'port{}'.format(p),
             'tcp': '{}:{}'.format(address, p),
             'Interval': '30s',
             'timeout': '4s'})
    return checks


def _cmd(cmd):
    """Execute cmd on the shell"""
    return subprocess.call(cmd, shell=True)
